#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "fmm_mpi.h"
#include "heap.h"
#include "array_mpi.h"


#define BAND 0
#define OUTSIDE -1
#define KNOWN n

//#define DEBUG
/*
 * number of loops
 */
int n;

/*
 * largest_solution is the largest value in the array
 * rollbacksmallest is the smallest value of received border values, which dictates the rollback number
 */
float largest_solution, rollbacksmallest;

/*
 * the n value which one should rollback to
 */
int rollbackn;
/*
 * mpi datatype for sending border points
 */
MPI_Datatype mpi_element_struct;
/*
 * Used for debug output
 */
float valuemax=0;
float valuemin=0;
/*
 * array containing the working status of each node
 */
int* working;

/*
 * Add a point to the heap
 */
void addToHeap(FmmData* data, int px, int py, int pz){
	Element *temp;
	#ifdef DEBUG
	printf("%d: adding %d, %d, %d to heap,heap size is %d maxsize if %d\n",cartrank,px,py,pz,data->heap->heapsize,data->heap->maxsize);
	#endif
	temp= malloc(sizeof(Element));
	temp->value = data->timearray[GETLINDEX(px,py,pz)];
	temp->x = px;
	temp->y = py;
	temp->z = pz;
	heapInsert(data->heap,temp);
	#ifdef DEBUG
	printf("%d: inserted %d, %d, %d to heap,heap size is %d\n",cartrank,px,py,pz,data->heap->heapsize);
	#endif

}

/*
 * initialize the FMM
 * velarray is the velocity field
 * posx,y,z is position of the starting point
 * x,y,z is the size of the array, most are read from array_mpi.h
 */
FmmData* initFMM(float* velarray, int posx,int posy, int posz, int x, int y, int z){
 	working = malloc(sizeof(int) * size);
	/* init datatypes*/
 	MPI_Element e;		
	MPI_Datatype type[5] = { MPI_FLOAT, MPI_INT, MPI_INT, MPI_INT, MPI_INT };
	int blocklen[5] = { 1, 1, 1, 1 , 1};
	MPI_Aint disp[5];
	disp[0] = 0;
	disp[1] = sizeof(float);
	disp[2] = sizeof(int) + disp[1];
	disp[3] = sizeof(int) + disp[2];
	disp[4] = sizeof(int) + disp[3];
	MPI_Type_create_struct(5, blocklen, disp, type, &mpi_element_struct);
	MPI_Type_commit(&mpi_element_struct);	
	/* end init datatypes*/
	#ifdef DEBUG
	printf("%d: done init mpi datatypes\n",cartrank);
	#endif
	int send = 0;
	FmmData* data;
	data = malloc(sizeof(FmmData));
	if(data == 0){
		printf("%d: Failed to allocate memory, exiting\n",cartrank);
		exit(1);
	}
	
	data->bandarray = malloc(sizeof(int)*LOCALARRAYSIZE);
	if(data->bandarray == 0){
		printf("%d: Failed to allocate memory, exiting\n",cartrank);
		exit(1);
	}
	data->timearray = malloc(sizeof(float)*LOCALARRAYSIZE);
	if(data->timearray == 0){
		printf("%d: Failed to allocate memory, exiting\n",cartrank);
		exit(1);
	}

	data->velocityarray = velarray;
	#ifdef DEBUG
	printf("%d: initializing heap\n",cartrank);
	#endif
	data->heap = initHeap(LOCALARRAYSIZE);
	data->x = x;
	data->y = y;
	data->z = z;
	data->posx = posx;
	data->posy = posy;
	data->posz = posz;

	#ifdef DEBUG
	printf("%d: clearing memory bandarray %p size %d\n",cartrank,data->bandarray,sizeof(int)*LOCALARRAYSIZE);
	fflush(stdout);
	#endif
	/*
	 * setting the band array to outside
	 */
	memset(data->bandarray,OUTSIDE,sizeof(int)*LOCALARRAYSIZE);
	#ifdef DEBUG
	printf("%d: Heap ok, clearing memory timearray %p size %d\n",cartrank,data->timearray,sizeof(float)*LOCALARRAYSIZE);
	#endif
	/*
	 * zeroing the arrival time array
	 */
	int i = 0;
	for(i=0;i<LOCALARRAYSIZE;i++)
		data->timearray[i] = 0;
	//memset(data->timearray,0,sizeof(float)*LOCALARRAYSIZE);
	//bzero(data->timearray,sizeof(float)*LOCALARRAYSIZE);
	data->sentarray = malloc(sizeof(float)*LOCALARRAYSIZE);
	bzero(data->sentarray,sizeof(float)*LOCALARRAYSIZE);
	#ifdef DEBUG
	printf("%d: done allocating memory, setting starting point\n",cartrank);
	fflush(stdout);
	#endif
	/*
	 * inserting the starting point on the correct node and add it to the heap / narrow band
	 */
	if(getDest(posx,posy,posz) == cartrank){
		int *cr = getLocalCord(posx,posy,posz);
		// insert starting point
		data->timearray[GETLINDEX(cr[0],cr[1],cr[2])] = 0;
		data->bandarray[GETLINDEX(cr[0],cr[1],cr[2])] = BAND;
		//printf("%d: sat pos %d %d %d, as known\n",cartrank,cr[0],cr[1],cr[2]);
		Element* element;
		element = malloc(sizeof(Element));
		element->x = cr[0];
		element->y = cr[1];
		element->z = cr[2];
		element->value = data->timearray[GETLINDEX(cr[0],cr[1],cr[2])];
		heapInsert(data->heap,element);
		send = 1;
	}
	return data;
}
/*
 * return min
 */
float min(float per, float truls){
	if(per > truls)
		return truls;
	else return per;
}
/*
 * return max
 */
float max(float per, float truls){
	if(per > truls)
		return per;
	else return truls;
}

/*
 * prints xy plane from a float array
 */
void printFloatArray(int sizex, int sizey, int z,float* array){
	int i,j;
	printf("\n Printing matrix\n");
	for(i=0;i<sizex;i++){
		for(j=0;j<sizey;j++){
			printf(" %8.2f ",array[GETLINDEX(i,j,z)]);
		}
		printf("\n");
	}
}

/*
 * calculate the arrival time for a point x,y,z
 */
float calcDistance(FmmData* data, int x, int y, int z){
	float sol;
	sol = BIGFLOAT;
	if(data->bandarray[GETLINDEX(x+1,y,z)] > BAND){
		sol = min(data->timearray[GETLINDEX(x+1,y,z)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for x+1\n",cartrank,data->timearray[GETLINDEX(x+1,y,z)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x-1,y,z)] > BAND){
		sol = min(data->timearray[GETLINDEX(x-1,y,z)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for x-1\n",cartrank,data->timearray[GETLINDEX(x-1,y,z)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x,y+1,z)] > BAND){
		sol = min(data->timearray[GETLINDEX(x,y+1,z)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for y+1\n",cartrank,data->timearray[GETLINDEX(x,y+1,z)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x,y-1,z)] > BAND){
		sol = min(data->timearray[GETLINDEX(x,y-1,z)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for y-1\n",cartrank,data->timearray[GETLINDEX(x,y-1,z)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x,y,z+1)] > BAND){
		sol = min(data->timearray[GETLINDEX(x,y,z+1)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for z+1\n",cartrank,data->timearray[GETLINDEX(x,y,z+1)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x,y,z-1)] > BAND){
		sol = min(data->timearray[GETLINDEX(x,y,z-1)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for z-1\n",cartrank,data->timearray[GETLINDEX(x,y,z-1)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	
	return sol;
}

/*
 * calculate a new point the the arrival time array if its inside the array and not KNOWN, if its outside add it to the heap
 */
void calcElement(FmmData* data, int px, int py, int pz){
        int add = 0;
        float sol;
        Element *temp;
        /*
         * Check if the point is inside the array*/
        if(px >= 0 && px < data->x && py >= 0 && py < data->y  && pz >= 0 && pz < data->z){
        /* Make sure the point is not KNOWN */
        if(data->bandarray[GETLINDEX(px,py,pz)] <= BAND){
                sol = calcDistance(data,px,py,pz);
                /* Check if the number is not smaller than the one we calculated, when we exit, should never happend in serial version*/
                if(data->timearray[GETLINDEX(px,py,pz)] != 0 && data->timearray[GETLINDEX(px,py,pz)] <= sol){
                        return;
                }
                /* If the point is OUSIDE add it to the heap/narrowband */
                if(data->bandarray[GETLINDEX(px,py,pz)] == OUTSIDE) {
                        data->bandarray[GETLINDEX(px,py,pz)] = BAND;
                        add = 1;
                }

                #ifdef DEBUG
                printf("%d: sol is %f, for %d %d %d\n",cartrank,sol,px,py,pz);
                #endif
                /* store max and min for debug purposes */
                #ifdef DEBUG
                if(sol > valuemax && sol != BIGFLOAT){
                        valuemax=sol;
                }
                if(sol < valuemin){
                        valuemin = sol;
                }
                #endif
                /* set the new arrivaltime */
                data->timearray[GETLINDEX(px,py,pz)] = sol;
                #ifdef DEBUG
                printf("%d: x %d , y %d, z %d, k[i] %d l[i] %d m[i] %d, value %f \n",cartrank,data->x, data->y, data->z,px,py,pz, data->timearray[GETLINDEX(px,py,pz)]);
                #endif
                /* add it to the narrowband if it should be added */
                if(add){
                        addToHeap(data,px,py,pz);
                        }
                }
        }
}

//remove ???
void checkforchange(FmmData* data,int px, int py, int pz){
	float sol = BIGFLOAT;
	int i;
	int o[6] = {px-1,px,px+1,px,px,px};
	int l[6] = {py,py-1, py,py+1,py,py};
	int m[6] = {pz,pz,pz,pz,pz-1,pz+1};
	for(i=0;i<6;i++){
		if(o[i] >= 0 && o[i] < data->x && l[i] >= 0 && l[i] < data->y  && m[i] >= 0 && m[i] < data->z && data->bandarray[GETLINDEX(o[i],l[i],m[i])] > BAND){
			
			#ifdef DEBUG
			printf("%d: checking for rollback %d %d %d \n",cartrank,o[i],l[i],m[i]);
			#endif
			sol = calcDistance(data,px,py,pz);
			
			if(sol < data->timearray[GETLINDEX(o[i],l[i],m[i])] && data->bandarray[GETLINDEX(o[i],l[i],m[i])] > BAND){
				// new value is smaller lets add this to our heap

				#ifdef DEBUG
				printf("%d: rolling back %d %d %d old value %f new value %f\n",cartrank,o[i],l[i],m[i],data->timearray[GETLINDEX(o[i],l[i],m[i])],sol);
				#endif
				data->timearray[GETLINDEX(o[i],l[i],m[i])] = sol;
				data->bandarray[GETLINDEX(o[i],l[i],m[i])] = BAND;
				addToHeap(data,o[i],l[i],m[i]);
				checkforchange(data,o[i],l[i],m[i]);
			}
		}
	}
}

/*
 * Add a new element to the array
 */
void addElement(FmmData* data,MPI_Element e){
	int* cr;
	int i,j,k;
	cr = getLocalCord(e.x,e.y,e.z);

	#ifdef DEBUG
	printf("%d: adding element %d %d %d, to local %d %d %d\n",cartrank,e.x,e.y,e.z,cr[0],cr[1],cr[2]);
	#endif
	/* 
	 * checking if we have added it before
	 * Should be always no since we don't send the same value multiple times
	 */
	if(data->timearray[GETLINDEX(cr[0],cr[1],cr[2])] == e.value){
		#ifdef DEBUG
		printf("%d: alreaddy added %d %d %d\n",cartrank,e.x,e.y,e.z);
		#endif
		return;
	}
	data->timearray[GETLINDEX(cr[0],cr[1],cr[2])] = e.value;
	data->bandarray[GETLINDEX(cr[0],cr[1],cr[2])] = e.n;
	int o[6] = {cr[0]-1,cr[0],cr[0]+1,cr[0],cr[0],cr[0]};
	int l[6] = {cr[1],cr[1]-1, cr[1],cr[1]+1,cr[1],cr[1]};
	int m[6] = {cr[2],cr[2],cr[2],cr[2],cr[2]-1,cr[2]+1};
	/*
	 * recalculate all neighbours
	 */
	for(i=0;i<6;i++){
		
		#ifdef DEBUG
		printf("%d: calc element %d %d %d\n",cartrank,o[i],l[i],m[i]);
		#endif
		calcElement(data, o[i], l[i], m[i]);
	}
	/* 
	 * set rollback values, so we can check if a rollback is necesarry
	 */
	/*if(largest_solution > e.value && rollbacksmallest > e.value){	
	rollbacksmallest = e.value;
	rollbackn = e.n;
	}*/
}

/*
 * rollback all values above input value
 */
void rollback(FmmData* data, float value){
	int i,j,k;	
	for(i=0;i<local_x;i++)
		for(j=0;j<local_y;j++)
			for(k=0;k<local_z;k++){
				if(data->bandarray[GETLINDEX(i,j,k)] > BAND && data->timearray[GETLINDEX(i,j,k)] > value){
					
					data->bandarray[GETLINDEX(i,j,k)] = BAND;
					addToHeap(data,i,j,k);
				}
			}
}

/*
 * send new values and check for incoming border values
 */
void sendRecvBorderChanges(FmmData* data,int x, int y, int z,int send){
	MPI_Element e;
	int run= 1;
	int *cr;
	int reast,rwest,rnorth,rsouth,rabove,rbelow;
	int seast,swest,snorth,ssouth,sabove,sbelow;
	reast = rwest = rnorth = rsouth = rabove = rbelow = 0;
	seast = swest = snorth = ssouth = sabove= sbelow = 0;
	/* 
	 * if we are to send a value
	 */
	if(send){
		/*
		 * see where we have to send the value
		 */
		if(x == 0){
			snorth = 1;
		}
		if(x == local_x-1){
			ssouth = 1;
		}
		if(y == 0){
			swest = 1;
			}
		if(y == local_y-1){
			seast = 1;
		}
		if(z == 0){
			sbelow = 1;
		}
		if(z == local_z-1){
			sabove = 1;
		}
		#ifdef DEBUG
		printf("%d: %d %d %d sending to north %d %d south %d  %d west %d %d east %d %d below %d %d above %d %d\n",cartrank,x,y,z,snorth,north,ssouth,south,swest,west,seast,east,sbelow,below,sabove,above);
		#endif
	
	
	if(snorth || ssouth || swest || seast || sabove || sbelow){
		cr = getGlobalCord(x,y,z);
		#ifdef DEBUG
		printf("%d: sending element at %d %d %d gave global coord %d %d %d\n",cartrank,x,y,z,cr[0],cr[1],cr[2]);
		#endif
		e.x = cr[0];
		e.y = cr[1];
		e.z = cr[2];
		e.value = data->timearray[GETLINDEX(x,y,z)];
		data->sentarray[GETLINDEX(x,y,z)] = e.value;
		e.n = data->bandarray[GETLINDEX(x,y,z)];
		
		if(snorth){
			MPI_Send(&e,1,mpi_element_struct,north,1,gridcomm);
		}
		if(ssouth){
			MPI_Send(&e,1,mpi_element_struct,south,1,gridcomm);
		}
		if(swest){
			MPI_Send(&e,1,mpi_element_struct,west,1,gridcomm);
		}
		if(seast){
			MPI_Send(&e,1,mpi_element_struct,east,1,gridcomm);
		}
		if(sabove){
			MPI_Send(&e,1,mpi_element_struct,above,1,gridcomm);
		}
		if(sbelow){
			MPI_Send(&e,1,mpi_element_struct,below,1,gridcomm);
		}
	}
	}
	rollbacksmallest = BIGFLOAT;
	rollbackn = 0;
	/*
	 * a loop to receive all incoming border values
	 */
	while(run){
	MPI_Iprobe(MPI_ANY_SOURCE,1,gridcomm,&run,MPI_STATUS_IGNORE);
	if(run){	
		MPI_Iprobe(north,1,gridcomm,&rnorth,MPI_STATUS_IGNORE);
		MPI_Iprobe(south,1,gridcomm,&rsouth,MPI_STATUS_IGNORE);
		MPI_Iprobe(east,1,gridcomm,&reast,MPI_STATUS_IGNORE);
		MPI_Iprobe(west,1,gridcomm,&rwest,MPI_STATUS_IGNORE);
		MPI_Iprobe(above,1,gridcomm,&rabove,MPI_STATUS_IGNORE);
		MPI_Iprobe(below,1,gridcomm,&rbelow,MPI_STATUS_IGNORE);
		if(rnorth || rsouth || rwest || reast || rabove || rbelow){
			
			if(rnorth){
				MPI_Recv(&e,1,mpi_element_struct,north,1,gridcomm,MPI_STATUS_IGNORE);
				addElement(data,e);
			}
			if(rsouth){
				MPI_Recv(&e,1,mpi_element_struct,south,1,gridcomm,MPI_STATUS_IGNORE);
				addElement(data,e);
			}
			if(rwest){
				MPI_Recv(&e,1,mpi_element_struct,west,1,gridcomm,MPI_STATUS_IGNORE);
				addElement(data,e);
			}
			if(reast){
				MPI_Recv(&e,1,mpi_element_struct,east,1,gridcomm,MPI_STATUS_IGNORE);
				addElement(data,e);
			}
			if(rabove){
				MPI_Recv(&e,1,mpi_element_struct,above,1,gridcomm,MPI_STATUS_IGNORE);
				addElement(data,e);
			}
			if(rbelow){
				MPI_Recv(&e,1,mpi_element_struct,below,1,gridcomm,MPI_STATUS_IGNORE);
				addElement(data,e);
			}
		}
	}
	}
	/* 
	 * rollback if necessary
	 */
	if(rollbackn){
		rollback(data,rollbacksmallest);
		n = rollbackn;
		rollbackn = 0;
		rollbacksmallest = BIGFLOAT;
	}
}

/*
 * old synchroneous border exhcange
 */
/*
void sendRecvBorderChanges(FmmData* data,int x, int y, int z,int send){
	MPI_Element e;
	
	int *cr;
	int reast,rwest,rnorth,rsouth,rabove,rbelow;
	int seast,swest,snorth,ssouth,sabove,sbelow;
	reast = rwest = rnorth = rsouth = rabove = rbelow = 0;
	seast = swest = snorth = ssouth = sabove= sbelow = 0;
	if(send){
		if(x == 0){
			snorth = 1;
		}
		if(x == local_x-1){
			ssouth = 1;
		}
		if(y == 0){
			swest = 1;
			}
		if(y == local_y-1){
			seast = 1;
		}
		if(z == 0){
			sbelow = 1;
		}
		if(z == local_z-1){
			sabove = 1;
		}
		#ifdef DEBUG
		printf("%d: %d %d %d sending to north %d %d south %d  %d west %d %d east %d %d below %d %d above %d %d\n",cartrank,x,y,z,snorth,north,ssouth,south,swest,west,seast,east,sbelow,below,sabove,above);
		#endif
	}
	
	MPI_Send(&swest,1,MPI_INT,west,0,gridcomm);
	MPI_Recv(&reast,1,MPI_INT,east,0,gridcomm,MPI_STATUS_IGNORE);
	MPI_Send(&seast,1,MPI_INT,east,0,gridcomm);
	MPI_Recv(&rwest,1,MPI_INT,west,0,gridcomm,MPI_STATUS_IGNORE);
		
	MPI_Send(&ssouth,1,MPI_INT,south,0,gridcomm);
	MPI_Recv(&rnorth,1,MPI_INT,north,0,gridcomm,MPI_STATUS_IGNORE);
	MPI_Send(&snorth,1,MPI_INT,north,0,gridcomm);
	MPI_Recv(&rsouth,1,MPI_INT,south,0,gridcomm,MPI_STATUS_IGNORE);
	
	MPI_Send(&sabove,1,MPI_INT,above,0,gridcomm);
	MPI_Recv(&rbelow,1,MPI_INT,below,0,gridcomm,MPI_STATUS_IGNORE);
	MPI_Send(&sbelow,1,MPI_INT,below,0,gridcomm);
	MPI_Recv(&rabove,1,MPI_INT,above,0,gridcomm,MPI_STATUS_IGNORE);
 		
	if(snorth || ssouth || swest || seast || sabove || sbelow){
		cr = getGlobalCord(x,y,z);
		
		#ifdef DEBUG
		printf("%d: sending element at %d %d %d gave global coord %d %d %d\n",cartrank,x,y,z,cr[0],cr[1],cr[2]);
		#endif
		e.x = cr[0];
		e.y = cr[1];
		e.z = cr[2];
		e.value = data->timearray[GETLINDEX(x,y,z)];
		e.n = data->bandarray[GETLINDEX(x,y,z)];
		
		if(snorth){
			MPI_Send(&e,1,mpi_element_struct,north,1,gridcomm);
		}
		if(ssouth){
			MPI_Send(&e,1,mpi_element_struct,south,1,gridcomm);
		}
		if(swest){
			MPI_Send(&e,1,mpi_element_struct,west,1,gridcomm);
		}
		if(seast){
			MPI_Send(&e,1,mpi_element_struct,east,1,gridcomm);
		}
		if(sabove){
			MPI_Send(&e,1,mpi_element_struct,above,1,gridcomm);
		}
		if(sbelow){
			MPI_Send(&e,1,mpi_element_struct,below,1,gridcomm);
		}
	}
	
	if(rnorth || rsouth || rwest || reast || rabove || rbelow){
		
		if(rnorth){
			MPI_Recv(&e,1,mpi_element_struct,north,1,gridcomm,MPI_STATUS_IGNORE);
			addElement(data,e);
		}
		if(rsouth){
			MPI_Recv(&e,1,mpi_element_struct,south,1,gridcomm,MPI_STATUS_IGNORE);
			addElement(data,e);
		}
		if(rwest){
			MPI_Recv(&e,1,mpi_element_struct,west,1,gridcomm,MPI_STATUS_IGNORE);
			addElement(data,e);
		}
		if(reast){
			MPI_Recv(&e,1,mpi_element_struct,east,1,gridcomm,MPI_STATUS_IGNORE);
			addElement(data,e);
		}
		if(rabove){
			MPI_Recv(&e,1,mpi_element_struct,above,1,gridcomm,MPI_STATUS_IGNORE);
			addElement(data,e);
		}
		if(rbelow){
			MPI_Recv(&e,1,mpi_element_struct,below,1,gridcomm,MPI_STATUS_IGNORE);
			addElement(data,e);
		}
	}
}
*/

/*
 * check if someone wants to update their working status
 */
void checkOthers(){
	int i;
	int flag;
	for(i = 0; i<size;i++){
		MPI_Iprobe(i,9,gridcomm,&flag,MPI_STATUS_IGNORE);
		if(flag){
			MPI_Recv(&working[i],1,MPI_INT,i,9,gridcomm,MPI_STATUS_IGNORE);
		}
	}
}
/*
 * notify others that my working status is changed
 */
void notifyOthers(int value){
	int i;
	for(i = 0; i<size;i++){
		MPI_Send(&value,1,MPI_INT,i,9,gridcomm);
	}
}

/*
 * Execute the FMM
 */
void executeFMM(FmmData* data){
	int add=0;
	int posx,posy,posz;
	int run = 1;
	int sendrun = 1;
	int senddata = 0;
	int end = 0;
	int sum = 0;
	int i;
	#ifdef DEBUG
	printf("data size is %d %d %d\n",data->x, data->y, data->z);
	#endif
	for(i = 0; i<size;i++){
		working[i] = 1;
	}	
	largest_solution = 0;
	n = 1;
	/* loop will run until all nodes are done */
	while(!end){
	/* working loop, will run until there are no more work to be done */
	while(run){


		if(heapGetMin(data->heap)){
		 
		 Element* e,*temp;
		
		 e = heapExtractMin(data->heap);
		 #ifdef DEBUG
		 printf("%d: setting %d %d %d to known\n",cartrank,e->x,e->y,e->z);
		 #endif
		 data->bandarray[GETLINDEX(e->x,e->y,e->z)] = KNOWN;
		 int k[6] = {e->x-1,e->x,e->x+1,e->x,e->x,e->x};
		 int l[6] = {e->y, e->y-1, e->y, e->y+1,e->y,e->y};
		 int m[6] = {e->z, e->z, e->z, e->z,e->z-1,e->z+1};
		 int i;
		 posx = e->x;
		 posy = e->y;
		 posz = e->z;
		 /*
		  * update largest_solution if this solution is the largest
		  */
		 if(data->timearray[GETLINDEX(posx,posy,posz)] > largest_solution){
		 	largest_solution = data->timearray[GETLINDEX(posx,posy,posz)];
		 }
		 /* check if this point has been sent before */
		 if(data->sentarray[GETLINDEX(posx,posy,posz)] == 0){
		 	senddata = 1;
		 }else if(data->sentarray[GETLINDEX(posx,posy,posz)] <= data->timearray[GETLINDEX(posx,posy,posz)]){
		 	senddata = 0;
		 }
		 
		 /*printFloatArray(data->x,data->y,-2,data->timearray);
		 printFloatArray(data->x,data->y,-1,data->timearray);
		 printFloatArray(data->x,data->y,0,data->timearray);
		 printFloatArray(data->x,data->y,1,data->timearray);
		 printFloatArray(data->x,data->y,2,data->timearray);*/
		 for(i=0;i<6;i++){
		 	
			if(k[i] >= 0 && k[i] < data->x && l[i] >= 0 && l[i] < data->y  && m[i] >= 0 && m[i] < data->z){
				calcElement(data, k[i], l[i], m[i]);	
			}
		 }
			
		#ifdef DEBUG

		 /*printFloatArray(data->x,data->y,-2,data->timearray);
		 printFloatArray(data->x,data->y,-1,data->timearray);
		 printFloatArray(data->x,data->y,0,data->timearray);
		 printFloatArray(data->x,data->y,1,data->timearray);
		 printFloatArray(data->x,data->y,2,data->timearray);*/
		#endif
		 free(e);

		 n++;
		}
		/* Send changes to border and look for incoming changes to the border */
		sendRecvBorderChanges(data,posx,posy,posz,senddata);	
		
		#ifdef DEBUG
		if(cartrank == 0|| cartrank == -2){printLocalArray(data->timearray);
		printLocalIntArray(data->bandarray);
		}
		#endif
		senddata = 0;
		/*Do we end the loop?*/
		if(heapGetMin(data->heap) == 0){
			run = 0;
		}else{
			run = 1;
		}
		checkOthers();
		#ifdef DEBUG
		if(cartrank == 0 && n%1000 == 0){
			printf("%d: reached n %d\n",cartrank,n);
		}
		#endif
	}
		/* see if border changes are coming */
		sendRecvBorderChanges(data,0,0,0,0);
		/* if we have work to do lets notify others and start to work again, if not let others know we are done */
		if(heapGetMin(data->heap) != 0){
			run = 1;
			// notify other i am still working
			notifyOthers(1);
		}else if(working[cartrank] == 1){
			// notify that i have stopped working
			notifyOthers(0);
		}
		checkOthers();
		sum = 0;
		for(i = 0; i<size;i++){
			sum += working[i];
		}	
		if(sum == 0){
			end = 1;
		}
	}
	#ifdef DEBUG
	printf("valuemax %f valuemin %f\n",valuemax,valuemin);
	#endif
}


/*
 * free the variables used in the FMM
 */
void freeFMM(FmmData* data){
	//free(data->timearray);
	free(data->bandarray);
	free(data);
}

