#include <stdlib.h>
#include <stdio.h>
#include "heap.h"

//#define DEBUG
/*
 * Initialize the heap with size as max heap size*/
Heap* initHeap(int size){
	Element **array;
	Heap* heap;
	#ifdef DEBUG
	printf("Ready to allocate heap\n");
	#endif
	heap = malloc(sizeof(Heap));
	array = malloc(sizeof(Element*)*size);
	
	#ifdef DEBUG
	printf("Allocated memory for heap\n");
	#endif
	heap->array = array;
	heap->maxsize = size;
	heap->heapsize=0;

	#ifdef DEBUG
	printf("returning heap\n");
	#endif
	return heap;
}

/*
 * A support function that checks that the heap is realy min sorted
 * Only used in testing if the heap works correctly, shoudn't be used in production code
 */
float checkHeapConsistensy(Heap* heap){
        int i;
        float min = 10000000000.0;
	if(heap->heapsize > 1){

	#ifdef DEBUG
	printf("checking consistency, size is %d\n", heap->heapsize);
	#endif
        for(i=1;i<=heap->heapsize;i++){
                if(min > heap->array[i]->value)
                min = heap->array[i]->value;
        }
	#ifdef DEBUG
	printf("root is %f, smallest is %f\n", heap->array[1]->value, min);
	#endif
	}
	return min;
}

/*
 * Returns the parten to a specific position in the heap
 */
int inline getParent(int pos){
	return pos>>1; //pos/2;
}
/*
 * Gets the left child of a specific position in the heap
 */
int inline getLeft(int pos){
	return pos <<1; // pos * 2
}
/*
 * Gets the right child of a specific position in the heap
 */
int inline getRight(int pos){
	return (pos <<1) +1; // pos*2+1
}

/*
 * Move one value to its correct position in the min sorted heap
 * Run once for each new element in the heap
 */
void MinHeapify(Heap* heap,int pos){
	int l,r;
	int smallest = 0;
	l = getLeft(pos);
	r = getRight(pos);
	if(l <= heap->heapsize && heap->array[l]->value < heap->array[pos]->value){
		smallest = l;
	}else{
		smallest = pos;
	}
	if(r<= heap->heapsize && heap->array[r]->value < heap->array[smallest]->value)
		smallest = r;
	if(smallest != pos){
		Element* temp;
		temp = heap->array[pos];
		heap->array[pos] = heap->array[smallest];
		heap->array[smallest] = temp;
		MinHeapify(heap,smallest);
	}
}

/*
 * return and remove the smallest element from the heap
 */
Element* heapExtractMin(Heap* heap){
	if(heap->heapsize<1) return NULL;
	Element* temp = heap->array[1];
	heap->array[1] = heap->array[heap->heapsize];
	heap->heapsize--;
	#ifdef DEBUG
	printf("New heapsize %d\n",heap->heapsize);
	#endif

	MinHeapify(heap,1);
	return temp;
}

/*
 * Return the smallest element from the heap
 * Does not remove anything from the heap
 */
Element* heapGetMin(Heap* heap){
	if(heap->heapsize<1) return NULL;
	Element* temp = heap->array[1];
	return temp;
}

/*
 * Add a new element to the heap
 * this function adds a element to a position and moves it up until its found its place in the heap.
 */
void heapIncreaseKey(Heap* heap,int pos, Element* key){
	Element* temp;
	//if(key->value > array[pos]->value){
	if(heap->array[pos] != 0){
		exit(2);
	}
	#ifdef DEBUG
	printf("Inserting into %d\n",pos);
	#endif
	heap->array[pos] = key;
	while (pos > 1 && heap->array[getParent(pos)]->value > heap->array[pos]->value){
		#ifdef DEBUG
		printf("chaging position between %d and %d\n",pos,getParent(pos));
		#endif
		temp = heap->array[pos];
		heap->array[pos] = heap->array[getParent(pos)];
		heap->array[getParent(pos)] = temp;
		pos = getParent(pos);
	}
	//checkHeapConsistensy(heap);
}

/*
 * add a new element to the heap
 */
void heapInsert(Heap* heap,Element* key){
	heap->heapsize++;
	if(heap->heapsize > heap->maxsize) 
	{
		printf("Exceeded heap max size of %d\n",heap->maxsize);
		exit(5);
	}
	#ifdef DEBUG
	printf("New heapsize %d array at %p\n",heap->heapsize,heap->array);
	#endif
	heap->array[heap->heapsize] = 0;
	heapIncreaseKey(heap,heap->heapsize,key);
}


