#include <stdio.h>
#include <stdlib.h>
#include "array_mpi.h"
#include "fileio.h"
#include <mpi.h>
#include <string.h>
#include "fmm_mpi.h"
#include "time.h"

#define MPIDEBUG 0
#define DEBUG 0

/* set that no dimensions sould be cyclic */
int periods[3] = {0,0,0};
/* set the number of dimensions to use in the cartesian node grid*/
int ndims = 3;

int div_x, div_y, div_z;

float* local_array;
float* global_array;
float* file_array;

/*
 * different mpi datatypes for exchanging the borders
 */
MPI_Datatype xz_plane;
MPI_Datatype zy_plane;
MPI_Datatype xy_plane;
MPI_Datatype y_column;
MPI_Datatype y_column_resized;
/*
 * A test function to check if the array dosen't contain a value
 */
int checkArray(int x, int y, int z,float value){
		int i,j,k;
		int rvalue = 1;
		for(i= 0; i<x;i++)
			for(j=0;j<y;j++)
				for(k=0;k<z;k++){
					if(local_array[GETLINDEX(i,j,k)] != value){
						rvalue = 0;
						return rvalue;
					}
				}
		return rvalue;
	}
/*
 * Check if a global array dosen't contain a specific value
 */
int checkGArray(int x, int y, int z,float value){
		int i,j,k;
		int rvalue = 1;
		for(i= 0; i< x;i++)
			for(j=0;j<y;j++)
				for(k=0;k<z;k++){
					if(global_array[GETINDEX(i,j,k)] != value){
						rvalue = 0;
						return rvalue;
					}
				}
		return rvalue;
	}

/*
 * Divide the global matrix into smaller matrixes for each node.
 * This function calculates the local dimensions
 */
void divide_matrix(){
	
	div_x = x/dims[0];
	div_y = y/dims[1];
	div_z = z/dims[2];
	local_x = div_x;
	local_y = div_y;
	local_z = div_z;
	if(local_x * dims[0] != x){
		if(coords[0] == dims[0]){
			local_x = (x-(local_x*dims[0])) + local_x;
		}
	}
	if(local_y * dims[1] != y){
		if(coords[1] == dims[1]){
			local_y = (y-(local_y*dims[1])) + local_y;
		}
	}
	if(local_z * dims[2] != z){
		if(coords[2] == dims[2]){
			local_z = (z-(local_z*dims[2])) + local_z;
		}
	}
	printf("%d: local_x %d local_y %d local_z %d coords %d,%d,%d \n",cartrank, local_x, local_y, local_z,coords[0],coords[1],coords[2]);
}

/*
 * initialize datatypes for border exchange
 */
void initMPIDatatypes(){
	MPI_Type_vector(local_y,local_z,local_z+2,MPI_FLOAT,&zy_plane);
	MPI_Type_commit(&zy_plane);
	
	MPI_Type_vector(local_x,local_z,(local_z+2)*(local_y+2),MPI_FLOAT,&xz_plane);
	MPI_Type_commit(&xz_plane);
	
	MPI_Type_vector(local_y,1,local_z+2,MPI_FLOAT,&y_column);
	MPI_Type_commit(&y_column);
 	MPI_Type_create_resized(y_column,0,(local_z+2)*(local_y+2)*sizeof(float),&y_column_resized);
	MPI_Type_vector(1,2,1,y_column_resized,&xy_plane);
	MPI_Type_commit(&xy_plane);
}
/*
 * Initialize the program allocating local matrixes and initializing data types
 */
void init(){
	divide_matrix();
	local_array = malloc(sizeof(float)*LOCALARRAYSIZE);
	if(local_array == 0){
		printf("Coudn't allocate enough memory for local_array\n");
		exit(1);
	}
	bzero(local_array,sizeof(float)*LOCALARRAYSIZE);
	initMPIDatatypes();
}
/*
 * a function for reading a file into each node, where the file contains a global array, each node will read its respective part into their local matrixes
 */
void scatterdata(char* filename){
	int i,j;
	int* cr;
	int value;
	char* errorstr;
	int reslen;
	FILE* f;
	int offset = 0;
	printf("%d: openeing file %s\n",cartrank,filename);
	f = fopen(filename,"rb");
	if(!f){
		printf("%d: unable to open file %d \n",cartrank,f);
		fflush(stdout);
		return;
	}
	if(DEBUG){
		printf("%d: opened file %d\n",cartrank,f);
	}
	for(i = 0; i<local_x;i++){
		for(j= 0; j< local_y; j++)
		{
			cr = getGlobalCord(i,j,0);
			offset = sizeof(float)*(cr[2]+cr[1]*z+cr[0]*z*y);
			fseek(f,offset,SEEK_SET);
			fread(&local_array[GETLINDEX(i,j,0)],sizeof(float),local_z,f);
			free(cr);
		}
	}
	fclose(f);

}

/*
 * gather all the local matrixes into a global matrix on node 0
 */
float* gatherdata(float* iarray){
	int i,j,k,dest,flag,r,t;
	float* farray;
	MPI_Request* requests;
	MPI_Status* status;
	requests = malloc(sizeof(MPI_Request)*local_y*local_x*2);
	status = malloc(sizeof(MPI_Status)*local_y*local_x*2);
	if(cartrank == 0){
		if(DEBUG){
			printf("%d:started gathering\n",cartrank);
		}
		farray = malloc(sizeof(float)*ARRAYSIZE);
		bzero(farray,sizeof(float)*ARRAYSIZE);
		
	}	
	if(cartrank != 0){
	for(i=0;i<local_x;i++)
		for(j= 0; j<local_y;j++){
			MPI_Send(&iarray[GETLINDEX(i,j,0)],local_z,MPI_FLOAT,0,i*local_y+j,gridcomm);//,&requests[(local_y*i)+j]);
		}
	
	if(DEBUG){
		printf("%d: Done sending \n",cartrank);
	}
	}
	if(cartrank==0){
		if(MPIDEBUG){
			printf("%d: starting setting recvs\n",cartrank);
		}
		for(r=0;r<x;r+=local_x)
			for(t=0; t<y;t+=local_y)
				for(k=0;k<dims[2];k++){
				
					dest = getDest(r,t,k*local_z);
					if(DEBUG){
						printf("%d: receiving from %d\n",cartrank,dest);
					}
					for(i=0;i<local_x;i++)
						for(j=0;j<local_y;j++){
						MPI_Irecv(&farray[GETINDEX(i+r,j+t,k*local_z)],local_z,MPI_FLOAT,dest,i*local_y+j,gridcomm,&requests[(local_y*local_x)+(i*local_y+j)]);
						if(dest == 0){
							MPI_Isend(&iarray[GETLINDEX(i,j,0)],local_z,MPI_FLOAT,0,i*local_y+j,gridcomm,&requests[(local_y*i)+j]);
						}
						}
					if(dest == 0){	
						MPI_Waitall(local_y*local_x*2,requests,status);
					}
					if(dest != 0){
					MPI_Waitall(local_y*local_x,&requests[local_y*local_x],status);
					}
				}
	}

	printf("%d: Done gathering \n",cartrank);
	return farray;
}


/*
 * check if the global array is the same as the array inside a file
 */
void checkData(char* filename){
	
	if(cartrank == 0){
	int i,j,k;		
		printf("%d: Reading file\n",cartrank);
		file_array = malloc(ARRAYSIZE*sizeof(float));
		bzero(file_array,ARRAYSIZE*sizeof(float));
		readfile(file_array,filename,x,y,z);
		
		printArray(file_array);
		printf("%d: checking data consistency %f\n",cartrank,file_array[GETINDEX(0,0,0)]);
		for(i=0;i<x;i++)
			for(j=0;j<y;j++)
				for(k=0;k<z;k++){
					if(file_array[GETINDEX(i,j,k)] != global_array[GETINDEX(i,j,k)]){
						printf("%d: error at %d %d %d file %lf global %lf\n",cartrank,i,j,k,file_array[GETINDEX(i,j,k)],global_array[GETINDEX(i,j,k)]);
					}
				}
	}
}

/*
 * Exchange borders
 */
void exchangeBorders(){


	// sending /recving north south
	MPI_Send(&local_array[GETLINDEX(0,0,0)],1,zy_plane,north,0,gridcomm);
	MPI_Recv(&local_array[GETLINDEX(local_x,0,0)],1,zy_plane,south,0,gridcomm,MPI_STATUS_IGNORE);
	MPI_Send(&local_array[GETLINDEX(local_x-1,0,0)],1,zy_plane,south,1,gridcomm);
	MPI_Recv(&local_array[GETLINDEX(-1,0,0)],1,zy_plane,north,1,gridcomm,MPI_STATUS_IGNORE);
	
	
	//sending/recving east, west
	MPI_Send(&local_array[GETLINDEX(0,0,0)],1,xz_plane,west,2,gridcomm);
	MPI_Recv(&local_array[GETLINDEX(0,local_y,0)],1,xz_plane,east,2,gridcomm,MPI_STATUS_IGNORE);
	MPI_Send(&local_array[GETLINDEX(0,local_y-1,0)],1,xz_plane,east,3,gridcomm);
	MPI_Recv(&local_array[GETLINDEX(0,-1,0)],1,xz_plane,west,3,gridcomm,MPI_STATUS_IGNORE);
	
	//sending/receving above, below
	MPI_Send(&local_array[GETLINDEX(0,0,local_z-1)],1,xy_plane,above,4,gridcomm);
	MPI_Recv(&local_array[GETLINDEX(0,0,-1)],1,xy_plane,below,4,gridcomm,MPI_STATUS_IGNORE);
	MPI_Send(&local_array[GETLINDEX(0,0,0)],1,xy_plane,below,5,gridcomm);
	MPI_Recv(&local_array[GETLINDEX(0,0,local_z)],1,xy_plane,above,5,gridcomm,MPI_STATUS_IGNORE);
}

/*
 * a test function for writing a file with values
 */
void writeafile(){
	FILE* f;
	f = fopen("/work/idarbo/per.conv","w");
	int i,j,k;
	float value;
	for(i=0;i<x;i++)
		for(j=0;j<y;j++)
			for(k=0;k<z;k++){
			//value = i+j+k;
			value = 1;
			if(j < 8 && k < 8){
				value = 9;
			}
			
			fwrite(&value,sizeof(float),1,f);
			}
	fclose(f);
}

int main(int argc, char** argv){
	float* array,*time1,*time2;
	int timeusec,timesec, timeusec2, timesec2,rtimesec, rtimeusec;
	FmmData* data;
        x = SIZEX;
        y = SIZEY;
        z = SIZEZ;
        int i, j, k;
        /* initialize MPI*/
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Dims_create(size,ndims,dims);

	MPI_Cart_create(MPI_COMM_WORLD,ndims, dims, periods, 0, &gridcomm);
	
	MPI_Cart_shift(gridcomm, 0, 1, &north, &south);

	MPI_Cart_shift(gridcomm, 1, 1, &west, &east);
	MPI_Cart_shift(gridcomm, 2, 1, &below, &above);

	MPI_Comm_rank(gridcomm, &cartrank);
	
	MPI_Cart_coords(gridcomm, cartrank, 3, coords);
	//if(cartrank ==0)
	//writeafile();
	/* init the program */
	init();

	if(MPIDEBUG){
		printf("%d: west %d east %d south %d north %d below %d above %d cartrank %d\n",rank,west,east,south,north,below,above,cartrank);
		fflush(stdout);
	}
	
	
	
	if(cartrank == 0){
		printf("%d: Init complete reading data from file\n",cartrank);
		fflush(stdout);
	}
	/* read data from file */
	//scatterdata(argv[1]);

	/* instead of reading set the array to 1.0 */
	for(i=-1;i<=local_x;i++)
		for(j=-1;j<=local_y;j++)
			for(k=-1;k<=local_z;k++)
				local_array[GETLINDEX(i,j,k)] = 1.0;
	//printLocalArray(local_array);
	//MPI_Barrier(gridcomm);
	if(cartrank == 0){
		printf("%d: Read data exchanging borders\n",cartrank);
		fflush(stdout);
	}
	/* exchange borders so border values will be correct in the velocity array */
	exchangeBorders();
	
	if(cartrank == 0){
		printf("%d: Initializing FMM\n",cartrank);
		fflush(stdout);
	}
	/* initialize the FMM */
	data = initFMM(local_array,x/2,y/2,z/2,local_x,local_y,local_z);
	
	if(cartrank == 0){
		printf("%d: Executing FMM\n",cartrank);
		fflush(stdout);
	}
	/* take timing and execute the FMM*/
	timeusec = getTimeInMicroseconds();
	timesec = getTimeInSeconds();
	executeFMM(data);
	timeusec2 = getTimeInMicroseconds();
	timesec2 = getTimeInSeconds();
	fixTime(timesec,timeusec,timesec2,timeusec2,&rtimesec,&rtimeusec);
	if(cartrank == 0){
		printf("%d: execute FMM took %d seconds and %d microseconds\n",cartrank,rtimesec,rtimeusec);
	}
	if(DEBUG){
		printLocalArray(data->timearray);
	}
	MPI_Barrier(gridcomm);	
	
	if(cartrank == 0){
		printf("%d: FMM done gathering data\n",cartrank);
	}
	/*gather if you can place all data on one node in memory, else write to file*/
	//global_array = gatherdata(data->timearray);
	
	//global_array = gatherdata(local_array);
	
	
	
	
	if(cartrank == 0){
		if(DEBUG){
        		//printArray(global_array);
		}
	}
	//checkData(argv[1]);

	printf("%d is done ending gracefully.\n",cartrank);
	MPI_Finalize();

}
