#include "array_mpi.h"
#include <stdlib.h>
#include <stdio.h>
/*
 * stores requests and status from non blocking border exchange
 */
MPI_Request request[24];
MPI_Status status[24];
/*
 * converts global coord to local coord
 */
int* getLocalCord(int xg, int yg, int zg){
        int *cr;
        cr = malloc(sizeof(int)*3);
        cr[0] = xg;
        cr[1] = yg;
        cr[2] = zg;
	cr[0] -= coords[0] * local_x;
	cr[1] -= coords[1] * local_y;
	cr[2] -= coords[2] * local_z;
        return cr;
}

/*
 * convert local coord to global coord
 */
int* getGlobalCord(int x, int y, int z){
        int *cr;
        cr = malloc(sizeof(int)*3);
        cr[0] = x+ local_x*coords[0];
        cr[1] = y+ local_y*coords[1];
        cr[2] = z+ local_z*coords[2];
        return cr;
        }

/*
 * gets the rank of the node that have input position
 */
int getDest(int x, int y, int z){
        int cr[3];
        int value;
        cr[0] = x/local_x;
        cr[1] = y/local_y;
        cr[2] = z/local_z;
        MPI_Cart_rank(gridcomm,cr,&value);
        return value;
}

/*
 * print the local float array to stdout
 */
void printLocalArray(float* array){
        int i,j,k;
        for(k=-1;k<=local_z;k++){
        printf("%d:array z = %d\n",cartrank,k);
                for(i=-1;i<=local_x;i++){
                        printf("%d: ",cartrank);
                        for(j=-1;j<=local_y;j++){
                                printf(" %8f ",array[GETLINDEX(i,j,k)]);
                        }
                        printf("\n");
                }
        printf("\n\n");
        }
}
/*
 * print a local integer array to stdout
 */
void printLocalIntArray(int* array){
        int i,j,k;
        for(k=-1;k<=local_z;k++){
        printf("%d:array z = %d\n",cartrank,k);
                for(i=-1;i<=local_x;i++){
                        printf("%d: ",cartrank);
                        for(j=-1;j<=local_y;j++){
                                printf(" %d ",array[GETLINDEX(i,j,k)]);
                        }
                        printf("\n");
                }
        printf("\n\n");
        }
}

/*
 * print the global array to stdout
 */
void printArray(float* array){
        int i,j,k;

        for(k=-1;k<=z;k++){
        printf("array z = %d\n",k);
                for(i=-1;i<=x;i++){
                        for(j=-1;j<=y;j++){
                                printf(" %8f ",array[GETINDEX(i,j,k)]);
                        }
                printf("\n");
                }
        printf("\n\n");
        }
}

/*
 * start non blocking exchange border communication
 */
void exchangeBorders(float* array){


        // sending /recving north south
        MPI_Isend(&array[GETLINDEX(0,0,0)],1,zy_plane,north,0,gridcomm,&request[0]);
        MPI_Irecv(&array[GETLINDEX(local_x,0,0)],1,zy_plane,south,0,gridcomm,&request[1]);//MPI_STATUS_IGNORE);
        MPI_Isend(&array[GETLINDEX(local_x-1,0,0)],1,zy_plane,south,1,gridcomm,&request[2]);
        MPI_Irecv(&array[GETLINDEX(-1,0,0)],1,zy_plane,north,1,gridcomm,&request[3]);//MPI_STATUS_IGNORE);


        //sending/recving east, west
        MPI_Isend(&array[GETLINDEX(0,0,0)],1,xz_plane,west,2,gridcomm,&request[4]);
        MPI_Irecv(&array[GETLINDEX(0,local_y,0)],1,xz_plane,east,2,gridcomm,&request[5]);//MPI_STATUS_IGNORE);
        MPI_Isend(&array[GETLINDEX(0,local_y-1,0)],1,xz_plane,east,3,gridcomm,&request[6]);
        MPI_Irecv(&array[GETLINDEX(0,-1,0)],1,xz_plane,west,3,gridcomm,&request[7]);//MPI_STATUS_IGNORE);

        //sending/receving above, below
        MPI_Isend(&array[GETLINDEX(0,0,local_z-1)],1,xy_plane,above,4,gridcomm,&request[8]);
        MPI_Irecv(&array[GETLINDEX(0,0,-1)],1,xy_plane,below,4,gridcomm,&request[9]);//MPI_STATUS_IGNORE);
        MPI_Isend(&array[GETLINDEX(0,0,0)],1,xy_plane,below,5,gridcomm,&request[10]);
        MPI_Irecv(&array[GETLINDEX(0,0,local_z)],1,xy_plane,above,5,gridcomm,&request[11]);//MPI_STATUS_IGNORE);
}

/*
 * initialize mpi datatypes
 */
void initMPIDatatypes(){
        MPI_Type_vector(local_y,local_z,local_z+2,MPI_FLOAT,&zy_plane);
        MPI_Type_commit(&zy_plane);

        MPI_Type_vector(local_x,local_z,(local_z+2)*(local_y+2),MPI_FLOAT,&xz_plane);
        MPI_Type_commit(&xz_plane);

        MPI_Type_vector(local_y,1,local_z+2,MPI_FLOAT,&y_column);
        MPI_Type_commit(&y_column);
        MPI_Type_create_resized(y_column,0,(local_z+2)*(local_y+2)*sizeof(float),&y_column_resized);
        MPI_Type_vector(1,2,1,y_column_resized,&xy_plane);
        MPI_Type_commit(&xy_plane);
}

/*
 * start exchanging integer border 
 */
void int_exchangeBorders(int* array){


        // sending /recving north south
        MPI_Isend(&array[GETLINDEX(0,0,0)],1,int_zy_plane,north,6,gridcomm,&request[12]);
        MPI_Irecv(&array[GETLINDEX(local_x,0,0)],1,int_zy_plane,south,6,gridcomm,&request[13]);//MPI_STATUS_IGNORE);
        MPI_Isend(&array[GETLINDEX(local_x-1,0,0)],1,int_zy_plane,south,7,gridcomm,&request[14]);
        MPI_Irecv(&array[GETLINDEX(-1,0,0)],1,int_zy_plane,north,7,gridcomm,&request[15]);//MPI_STATUS_IGNORE);


        //printf("%d: sending to %d reciving from %d\n",cartrank,west,east);
        //sending/recving east, west


        MPI_Isend(&array[GETLINDEX(0,0,0)],1,int_xz_plane,west,8,gridcomm,&request[16]);
        MPI_Irecv(&array[GETLINDEX(0,local_y,0)],1,int_xz_plane,east,8,gridcomm,&request[17]);//MPI_STATUS_IGNORE);
        MPI_Isend(&array[GETLINDEX(0,local_y-1,0)],1,int_xz_plane,east,9,gridcomm,&request[18]);
        MPI_Irecv(&array[GETLINDEX(0,-1,0)],1,int_xz_plane,west,9,gridcomm,&request[19]);//MPI_STATUS_IGNORE);

        //sending/receving above, below

        MPI_Isend(&array[GETLINDEX(0,0,local_z-1)],1,int_xy_plane,above,10,gridcomm,&request[20]);
        MPI_Irecv(&array[GETLINDEX(0,0,-1)],1,int_xy_plane,below,10,gridcomm,&request[21]);//MPI_STATUS_IGNORE);
        MPI_Isend(&array[GETLINDEX(0,0,0)],1,int_xy_plane,below,11,gridcomm,&request[22]);
        MPI_Irecv(&array[GETLINDEX(0,0,local_z)],1,int_xy_plane,above,11,gridcomm,&request[23]);//MPI_STATUS_IGNORE);
}

/*
 * initialise integer mpi datatypes
 */
void int_initMPIDatatypes(){
        MPI_Type_vector(local_y,local_z,local_z+2,MPI_INT,&int_zy_plane);
        MPI_Type_commit(&int_zy_plane);

        MPI_Type_vector(local_x,local_z,(local_z+2)*(local_y+2),MPI_INT,&int_xz_plane);
        MPI_Type_commit(&int_xz_plane);

        MPI_Type_vector(local_y,1,local_z+2,MPI_INT,&int_y_column);
        MPI_Type_commit(&int_y_column);
        MPI_Type_create_resized(int_y_column,0,(local_z+2)*(local_y+2)*sizeof(int),&int_y_column_resized);
        MPI_Type_vector(1,2,1,int_y_column_resized,&int_xy_plane);
        MPI_Type_commit(&int_xy_plane);
}

/*
 * Wait for int border exchange to finish
 */
void wait_exchange_int(){
	MPI_Waitall(12,&request[12],&status[12]);
}
/*
 * wait for float border exchange to finish
 */
void wait_exchange_float(){
	MPI_Waitall(12,request,status);
}

/*
 * wait for all border exchanges to finish
 */
void wait_exchange_all(){
	MPI_Waitall(24,request,status);
}

