#ifndef ARRAY_H
#define ARRAY_H

#include "mpi.h"
/*
 * Defines the size of the array
 */
#define SIZEX 448//448//8//448
#define SIZEY 704//704//4//704
#define SIZEZ 1216//1216//4//1216 
/*
 * Calculate the strides for global array
 */
#define stride0 ((SIZEZ+2)*(SIZEY+2))
#define stride1 (SIZEZ+2)
#define stride2 1
/*
 * Gets the index for a position in the global array
 */
#define GETINDEX(i,j,k) stride0*(i+1) + stride1*(j+1) + stride2*(k+1)
/*
 * Get the index for a position in the local array
 */
#define GETLINDEX(i,j,k) ((local_z+2)*(local_y+2)) *(i+1) + ((local_z+2)*(j+1)) + k+1
/*
 * A large float number, should be larger than anything you calculate
 */
#define BIGFLOAT 10000000.0

/*
 * The size of each array, for mallocing memory
 */
#define ARRAYSIZE (SIZEX+2)*(SIZEY+2)*(SIZEZ+2)
#define LOCALARRAYSIZE (local_x+2)*(local_y+2)*(local_z+2)

/*
 * Rank, MPI rank
 * cartrank, rank in the cartesian grid
 * size, number of nodes used
 */
int rank,cartrank, size;
/*
 * Ranks of nodes that are above,below,west,east,north,south
 */
int above, below, west, east, north ,south;
/*
 * Number of nodes in each dimension
 */
int dims[3];
/*
 * Different datatypes for local arrays
 */
MPI_Datatype xz_plane;
MPI_Datatype zy_plane;
MPI_Datatype xy_plane;
MPI_Datatype y_column;
MPI_Datatype y_column_resized;

MPI_Datatype int_xz_plane;
MPI_Datatype int_zy_plane;
MPI_Datatype int_xy_plane;
MPI_Datatype int_y_column;
MPI_Datatype int_y_column_resized;
/*
 * communicator for the cartesian grid
 */
MPI_Comm gridcomm;
/*
 * MY coordinates in the cartesian node grid
 */
int coords[3];
/*
 * size in each dimension of the global array
 */
int x,y,z;
/*
 * Size in each dimension of the local array
 */
int local_x,local_y,local_z;

/*
 * Print the values of a global array to stdout
 */
void printArray(float* array);
/*
 * Print the local array to stdout
 */
void printLocalArray(float* array);
/*
 * print a local int array to stdout
 */
void printLocalIntArray(int* array);
/*
 * Get what node a global position resides in
 */
int getDest(int x, int y, int z);
/*
 * Get global coordinates from local coordinates
 */
int* getGlobalCord(int x, int y, int z);
/*
 * Get local coordinates from global coordinates
 */
int* getLocalCord(int x, int y, int z);
/*
 * Start exchanging of borders
 */
void exchangeBorders(float* array);
/*
 * Initializa mpi data types
 */
void initMPIDatatypes();
/*
 * Start exchanging integer borders
 */
void int_exchangeBorders(int* array);
/*
 * Initialize integer datatypes
 */
void int_initMPIDatatypes();
/*
 * Wait for all int exchange borders
 */
void wait_exchange_int();
/*
 * wait for all float exchange borders
 */
void wait_exchange_float();
/*
 * Wait for all border exchanges
 */
void wait_exchange_all();

#endif
