#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "fmm_mpi.h"
#include "heap.h"
#include "array_mpi.h"


#define BAND 0
#define OUTSIDE -1
#define KNOWN n

//#define DEBUG
int n;

/*
 * Used for debug printing
 */
float valuemax=0;
float valuemin=0;

/*
 * Add a point to the heap
 */
void addToHeap(FmmData* data, int px, int py, int pz){
	Element *temp;
		
	#ifdef DEBUG
	printf("%d: adding %d, %d, %d to heap,heap size is %d maxsize if %d\n",cartrank,px,py,pz,data->heap->heapsize,data->heap->maxsize);
	#endif
	temp= malloc(sizeof(Element));
	temp->value = data->timearray[GETLINDEX(px,py,pz)];
	temp->x = px;
	temp->y = py;
	temp->z = pz;
	heapInsert(data->heap,temp);
	#ifdef DEBUG
	printf("%d: inserted %d, %d, %d to heap,heap size is %d\n",cartrank,px,py,pz,data->heap->heapsize);
	#endif
}

/*
 * initialize the FMM
 * velarray is the velocity field
 * posx,y,z is position of the starting point
 * x,y,z is the size of the array, most are read from array_mpi.h
 */
FmmData* initFMM(float* velarray, int posx,int posy, int posz, int x, int y, int z){
	/* init datatypes*/
	int_initMPIDatatypes();
	/* end init datatypes*/
	#ifdef DEBUG
	printf("%d: done init mpi datatypes\n",cartrank);
	#endif
	FmmData* data;
	data = malloc(sizeof(FmmData));
	if(data == 0){
		printf("%d: Failed to allocate memory, exiting\n",cartrank);
		exit(1);
	}
	
	data->bandarray = malloc(sizeof(int)*LOCALARRAYSIZE);
	if(data->bandarray == 0){
		printf("%d: Failed to allocate memory, exiting\n",cartrank);
		exit(1);
	}
	data->timearray = malloc(sizeof(float)*LOCALARRAYSIZE);
	if(data->timearray == 0){
		printf("%d: Failed to allocate memory, exiting\n",cartrank);
		exit(1);
	}

	data->velocityarray = velarray;
	#ifdef DEBUG
	printf("%d: initializing heap\n",cartrank);
	#endif
	data->heap = initHeap(LOCALARRAYSIZE);
	data->x = x;
	data->y = y;
	data->z = z;
	data->posx = posx;
	data->posy = posy;
	data->posz = posz;

	#ifdef DEBUG
	printf("%d: clearing memory bandarray %p size %d\n",cartrank,data->bandarray,sizeof(int)*LOCALARRAYSIZE);
	fflush(stdout);
	#endif
	/*
	 * Setting the band array to ouside
	 */
	memset(data->bandarray,OUTSIDE,sizeof(int)*LOCALARRAYSIZE);
	#ifdef DEBUG
	printf("%d: Heap ok, clearing memory timearray %p size %d\n",cartrank,data->timearray,sizeof(float)*LOCALARRAYSIZE);
	#endif
	/*
	 * zeroing the arrival time array
	 */
	int i = 0;
	for(i=0;i<LOCALARRAYSIZE;i++)
		data->timearray[i] = 0;
	//memset(data->timearray,0,sizeof(float)*LOCALARRAYSIZE);
	//bzero(data->timearray,sizeof(float)*LOCALARRAYSIZE);
	#ifdef DEBUG
	printf("%d: done allocating memory, setting starting point\n",cartrank);
	fflush(stdout);
	#endif
	/* 
	 * insert the starting point on the correct node and add it to the heap
	 */
	if(getDest(posx,posy,posz) == cartrank){
		int *cr = getLocalCord(posx,posy,posz);
		data->timearray[GETLINDEX(cr[0],cr[1],cr[2])] = 0;
		data->bandarray[GETLINDEX(cr[0],cr[1],cr[2])] = BAND;
		Element* element;
		element = malloc(sizeof(Element));
		element->x = cr[0];
		element->y = cr[1];
		element->z = cr[2];
		element->value = data->timearray[GETLINDEX(cr[0],cr[1],cr[2])];
		heapInsert(data->heap,element);
	}
	return data;
}

/*
 * return min
 */
float min(float per, float truls){
	if(per > truls)
		return truls;
	else return per;
}
/*
 * return max
 */
float max(float per, float truls){
	if(per > truls)
		return per;
	else return truls;
}

/*
 * prints a xy- plane from a float array
 */
void printFloatArray(int sizex, int sizey, int z,float* array){
	int i,j;
	printf("\n Printing matrix\n");
	for(i=0;i<sizex;i++){
		for(j=0;j<sizey;j++){
			printf(" %8.2f ",array[GETLINDEX(i,j,z)]);
		}
		printf("\n");
	}
}

/*
 * calculate the new arrivaltime for a point x,y,z
 */
float calcDistance(FmmData* data, int x, int y, int z){
	float sol;
	sol = BIGFLOAT;
	if(data->bandarray[GETLINDEX(x+1,y,z)] > BAND){
		sol = min(data->timearray[GETLINDEX(x+1,y,z)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for x+1\n",cartrank,data->timearray[GETLINDEX(x+1,y,z)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x-1,y,z)] > BAND){
		sol = min(data->timearray[GETLINDEX(x-1,y,z)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for x-1\n",cartrank,data->timearray[GETLINDEX(x-1,y,z)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x,y+1,z)] > BAND){
		sol = min(data->timearray[GETLINDEX(x,y+1,z)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for y+1\n",cartrank,data->timearray[GETLINDEX(x,y+1,z)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x,y-1,z)] > BAND){
		sol = min(data->timearray[GETLINDEX(x,y-1,z)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for y-1\n",cartrank,data->timearray[GETLINDEX(x,y-1,z)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x,y,z+1)] > BAND){
		sol = min(data->timearray[GETLINDEX(x,y,z+1)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for z+1\n",cartrank,data->timearray[GETLINDEX(x,y,z+1)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	if(data->bandarray[GETLINDEX(x,y,z-1)] > BAND){
		sol = min(data->timearray[GETLINDEX(x,y,z-1)] +1/data->velocityarray[GETLINDEX(x,y,z)],sol);
		#ifdef DEBUG
		printf("%d: sol is %f for z-1\n",cartrank,data->timearray[GETLINDEX(x,y,z-1)] +1/data->velocityarray[GETLINDEX(x,y,z)]);
		#endif
	}
	
	return sol;
}

/*
 * calculate a new point the the arrival time array if its inside the array and not KNOWN, if its outside add it to the heap
 */
void calcElement(FmmData* data, int px, int py, int pz){
	int add = 0;
	float sol;
	Element *temp;
	/*
	 * Check if the point is inside the array*/
	if(px >= 0 && px < data->x && py >= 0 && py < data->y  && pz >= 0 && pz < data->z){	
	/* Make sure the point is not KNOWN */
	if(data->bandarray[GETLINDEX(px,py,pz)] <= BAND){
		sol = calcDistance(data,px,py,pz);
		/* Check if the number is not smaller than the one we calculated, when we exit, should never happend in serial version*/
		if(data->timearray[GETLINDEX(px,py,pz)] != 0 && data->timearray[GETLINDEX(px,py,pz)] <= sol){
			return;
		}
		/* If the point is OUSIDE add it to the heap/narrowband */
		if(data->bandarray[GETLINDEX(px,py,pz)] == OUTSIDE) {
			data->bandarray[GETLINDEX(px,py,pz)] = BAND;
			add = 1;
		}

		#ifdef DEBUG
		printf("%d: sol is %f, for %d %d %d\n",cartrank,sol,px,py,pz);	
		#endif
		/* store max and min for debug purposes */
		#ifdef DEBUG
		if(sol > valuemax && sol != BIGFLOAT){
			valuemax=sol;
		}
		if(sol < valuemin){
			valuemin = sol;
		}
		#endif
		/* set the new arrivaltime */
		data->timearray[GETLINDEX(px,py,pz)] = sol;
		#ifdef DEBUG
		printf("%d: x %d , y %d, z %d, k[i] %d l[i] %d m[i] %d, value %f \n",cartrank,data->x, data->y, data->z,px,py,pz, data->timearray[GETLINDEX(px,py,pz)]);
		#endif
		/* add it to the narrowband if it should be added */
		if(add){
			addToHeap(data,px,py,pz);	
			}
		}
	}
}

/* 
 * reset the band array to outside for a new run
 * used when we receive new values
 */
void clearBand(FmmData* data){
	int i,j,k;
	int *cr;
	cr = getLocalCord(data->posx,data->posy,data->posz);
	for(i = 0;i<local_x;i++)
		for(j=0;j<local_y;j++)
			for(k=0;k<local_z;k++){
				if(i != cr[0] && j != cr[1] && k != cr[2]){
					data->bandarray[GETLINDEX(i,j,k)] = OUTSIDE;
				}
			}
}

/*
 * check if the new border values affect any values in the array,
 * if they do add them to the narrow band
 */
void addBordersToHeap(FmmData* data){
	int i,j,k;	
	float sol;
	i = local_x-1;
	for(j=0;j<local_y;j++)
		for(k=0;k<local_z;k++){
			if(data->bandarray[GETLINDEX(i+1,j,k)] > BAND && data->bandarray[GETLINDEX(i,j,k)] == OUTSIDE){
				sol = calcDistance(data,i,j,k);
				if(data->timearray[GETLINDEX(i,j,k)] == 0 || data->timearray[GETLINDEX(i,j,k)] > sol){
					data->timearray[GETLINDEX(i,j,k)] = sol;
					data->bandarray[GETLINDEX(i,j,k)] = BAND;
					//printf("%d: adding %d %d %d to heap\n",cartrank,i,j,k);
					addToHeap(data,i,j,k);
				}
			}
		}
	
	i = 0;
	for(j=0;j<local_y;j++)
		for(k=0;k<local_z;k++){
			if(data->bandarray[GETLINDEX(i-1,j,k)] > BAND && data->bandarray[GETLINDEX(i,j,k)] == OUTSIDE){
				sol = calcDistance(data,i,j,k);
				if(data->timearray[GETLINDEX(i,j,k)] == 0 || data->timearray[GETLINDEX(i,j,k)] > sol){
					data->timearray[GETLINDEX(i,j,k)] = sol;
					data->bandarray[GETLINDEX(i,j,k)] = BAND;
					//printf("%d: adding %d %d %d to heap\n",cartrank,i,j,k);
					addToHeap(data,i,j,k);
				}
			}
		}

	
	j = 0;
	for(i=0;i<local_x;i++)
		for(k=0;k<local_z;k++){
			if(data->bandarray[GETLINDEX(i,j-1,k)] > BAND && data->bandarray[GETLINDEX(i,j,k)] == OUTSIDE){
				sol = calcDistance(data,i,j,k);
				if(data->timearray[GETLINDEX(i,j,k)] == 0 || data->timearray[GETLINDEX(i,j,k)] > sol){
					data->timearray[GETLINDEX(i,j,k)] = sol;
					data->bandarray[GETLINDEX(i,j,k)] = BAND;
					//printf("%d: adding %d %d %d to heap\n",cartrank,i,j,k);
					addToHeap(data,i,j,k);
				}
			}
		}
	j = local_y-1;
	for(i=0;i<local_x;i++)
		for(k=0;k<local_z;k++){
			if(data->bandarray[GETLINDEX(i,j+1,k)] > BAND && data->bandarray[GETLINDEX(i,j,k)] == OUTSIDE){
				sol = calcDistance(data,i,j,k);
				if(data->timearray[GETLINDEX(i,j,k)] == 0 || data->timearray[GETLINDEX(i,j,k)] > sol){
					data->timearray[GETLINDEX(i,j,k)] = sol;
					data->bandarray[GETLINDEX(i,j,k)] = BAND;
					//printf("%d: adding %d %d %d to heap\n",cartrank,i,j,k);
					addToHeap(data,i,j,k);
				}
			}
		}
	k = local_z-1;
	for(i=0;i<local_x;i++)
		for(j=0;j<local_y;j++){
			if(data->bandarray[GETLINDEX(i,j,k+1)] > BAND && data->bandarray[GETLINDEX(i,j,k)] == OUTSIDE){
				sol = calcDistance(data,i,j,k);
				if(data->timearray[GETLINDEX(i,j,k)] == 0 || data->timearray[GETLINDEX(i,j,k)] > sol){
					data->timearray[GETLINDEX(i,j,k)] = sol;
					data->bandarray[GETLINDEX(i,j,k)] = BAND;
					//printf("%d: adding %d %d %d to heap\n",cartrank,i,j,k);
					addToHeap(data,i,j,k);
				}
			}
		}
	k = 0;
	for(i=0;i<local_x;i++)
		for(j=0;j<local_y;j++){
			if(data->bandarray[GETLINDEX(i,j,k-1)] > BAND && data->bandarray[GETLINDEX(i,j,k)] == OUTSIDE){
				sol = calcDistance(data,i,j,k);
				if(data->timearray[GETLINDEX(i,j,k)] == 0 || data->timearray[GETLINDEX(i,j,k)] > sol){
					data->timearray[GETLINDEX(i,j,k)] = sol;
					data->bandarray[GETLINDEX(i,j,k)] = BAND;
					//printf("%d: adding %d %d %d to heap\n",cartrank,i,j,k);
					addToHeap(data,i,j,k);
				}
			}
		}
}

/*
 * Execute the FMM
 */
void executeFMM(FmmData* data){
	int add=0;
	int posx,posy,posz;
	int run = 1;
	int sendrun = 1;
	int senddata = 0;
	int end = 0;
	int sum = 0;
	int i;
	#ifdef DEBUG
	printf("data size is %d %d %d\n",data->x, data->y, data->z);
	#endif
	n = 1;
	while(run){

		/*
		 * Run through heap
		 */
		while(heapGetMin(data->heap)!= 0){
		 
		 Element* e,*temp;
		
		 e = heapExtractMin(data->heap);
		 #ifdef DEBUG
		 printf("%d: setting %d %d %d to known\n",cartrank,e->x,e->y,e->z);
		 #endif
		 /*
		  * set the extracted point to KNOWN
		  */
		 data->bandarray[GETLINDEX(e->x,e->y,e->z)] = KNOWN;
		 int k[6] = {e->x-1,e->x,e->x+1,e->x,e->x,e->x};
		 int l[6] = {e->y, e->y-1, e->y, e->y+1,e->y,e->y};
		 int m[6] = {e->z, e->z, e->z, e->z,e->z-1,e->z+1};
		 int i;
		 posx = e->x;
		 posy = e->y;
		 posz = e->z;
		 
		 /*printFloatArray(data->x,data->y,-2,data->timearray);
		 printFloatArray(data->x,data->y,-1,data->timearray);
		 printFloatArray(data->x,data->y,0,data->timearray);
		 printFloatArray(data->x,data->y,1,data->timearray);
		 printFloatArray(data->x,data->y,2,data->timearray);*/
		 /*
		  * calculate neighbouring points
		  */
		 for(i=0;i<6;i++){
		 	
			if(k[i] >= 0 && k[i] < data->x && l[i] >= 0 && l[i] < data->y  && m[i] >= 0 && m[i] < data->z){
				calcElement(data, k[i], l[i], m[i]);	
			}
		 }
			
		#ifdef DEBUG

		 /*printFloatArray(data->x,data->y,-2,data->timearray);
		 printFloatArray(data->x,data->y,-1,data->timearray);
		 printFloatArray(data->x,data->y,0,data->timearray);
		 printFloatArray(data->x,data->y,1,data->timearray);
		 printFloatArray(data->x,data->y,2,data->timearray);*/
		#endif
		 free(e);
		}
		
		/*
		 * Exchange borders
		 */
		exchangeBorders(data->timearray);
		int_exchangeBorders(data->bandarray);
		wait_exchange_all();
		/*
		 * add Borders to heap if needed
		 */
		clearBand(data);
		addBordersToHeap(data);
		/*
		 * check if we are done
		 */
		if(heapGetMin(data->heap) != 0){
			sendrun = 1;
		}else {
			sendrun = 0;
		}
		//printf("%d: are we done %d\n",cartrank,sendrun);
		MPI_Allreduce(&sendrun,&run,1,MPI_INT,MPI_MAX,gridcomm);
	}
	#ifdef DEBUG
	printf("valuemax %f valuemin %f\n",valuemax,valuemin);
	//intLocalArray(data->timearray);
	#endif
}

/*
 * free the variables used in the FMM
 */
void freeFMM(FmmData* data){
	//free(data->timearray);
	free(data->bandarray);
	free(data);
}

